/****************************************************************************
** COPYRIGHT (C):    1996 Cay S. Horstmann. All Rights Reserved.
**
** NOTE TO STUDENTS: Do not attempt to study the contents of this file. You 
** can, and should, use the services provided in this file without knowing
** the highly technical details of the implementation.
**
** PROJECT:          Computing Concepts with C++ 
** FILE:             ccc_bgi.cpp
** PURPOSE:          Graphics window for BGI graphics
** VERSION           1.0
** PROGRAMMER:       Cay Horstmann (CSH)
**                   Jiaoyang Zhou
** START DATE:       2-19-94  (CSH)
** UPDATE HISTORY:
**
****************************************************************************/

/* IMPORT ******************************************************************/

#define EXPORT

#ifndef EXPORT

#ifdef CCC_USE_OLD_CLIBS
   #include <math.h>
   #include <string.h>
   #include <stdlib.h>
   #include <stdio.h>
#else
   #include <cmath>
   #include <cstring>
   #include <cstdlib>
   #include <cstdio>
#endif

#include <dos.h>
#include <bios.h>
#include <graphics.h>

#include <io.h>
#include <conio.h>

#endif
   // EXPORT
   
#include "ccc_bgi.h"

/* CLASSES *****************************************************************/

#ifndef EXPORT

class GraphicWindow;

class MouseEvent
{
public:
   enum Type { NOT_INSTALLED, LBUTTONDOWN, MOUSEMOVE, LBUTTONUP };
   MouseEvent(Type, double, double);

   Type type() const { return _type; }
   double x() const { return _x; }
   double y() const { return _y; }
private:
   Type _type;
   double _x;
   double _y;
};

/*-------------------------------------------------------------------------*/

class Mouse
{
public:
   Mouse(GraphicWindow& w);

   MouseEvent read(const GraphicWindow& gc);

   bool is_installed() const;

   int get_position(int&, int&) const;
   void show_cursor();
   void hide_cursor();
private:
   enum
   {  MOUSE_INTR = 0x33,
      RETF = 0xCF, // far return instruction
      SHOWCURSOR = 1,
      HIDECURSOR = 2,
      READMOUSE = 3,
   };
   bool _installed;
   GraphicWindow* ws;
};

/*-------------------------------------------------------------------------*/

class GraphicWindow
{
public:
   GraphicWindow();
   ~GraphicWindow();

   GraphicWindow& operator<<(Point p);
   GraphicWindow& operator<<(Circle c);
   GraphicWindow& operator<<(Line s);
   GraphicWindow& operator<<(Message t);

   string get_string(string prompt);
   int get_int(const string& out_string) 
   {  return atoi(get_string(out_string).c_str()); 
   }
   double get_double(const string& out_string) 
   {  return atof(get_string(out_string).c_str()); 
   }
   Point get_mouse(string prompt);

   void coord(double xmin, double ymin, double xmax, double ymax);
   void clear();

private:
   int user_to_disp_x(double x) const;
   int user_to_disp_y(double y) const;
   int user_to_disp_dx(double x) const;
   int user_to_disp_dy(double y) const;

   double disp_to_user_dx(int x) const;
   double disp_to_user_dy(int y) const;
   double disp_to_user_x(int x) const;
   double disp_to_user_y(int y) const;

   void point(double x, double y);
   void line(double xfrom, double yfrom, double xto, double yto);
   void ellipse(double x, double y, double ra, double rb);

   void text(string t, double x, double y);

   void statusline_prompt(string s);
   void statusline_echo(string s);
   void statusline_clear();

   GraphicWindow::GraphicWindow(const GraphicWindow&);
   GraphicWindow& operator=(const GraphicWindow&);

   enum { POINT_RADIUS = 2 };

   double _user_xmin;
   double _user_xmax;
   double _user_ymin;
   double _user_ymax;

   int _disp_xmax;
   int _disp_ymax;
   int _disp_xoff;
   int _disp_yoff;

   Mouse _mouse;
   string _prompt;   // current string in the graphics window
   string _echo;

friend class Mouse;
};

#endif
   // EXPORT

/* GLOBALS *****************************************************************/

EXPORT GraphicWindow cwin;

/* FUNCTIONS ***************************************************************/

int ccc_win_main();

int main()
{  return ccc_win_main();
}

/*-------------------------------------------------------------------------*/

Point GraphicWindow::get_mouse(string s)
{  Mouse m(*this);
   if (!m.is_installed())
   {  cerr << "get_mouse: Can't initialize mouse. Is the mouse driver loaded?\n";
      exit(EXIT_FAILURE);
   }

   statusline_prompt(s);
   while(true)
   {  int x,y;
      m.get_position(x, y);
      char buf[100];
      sprintf(buf, "(%f,%f)",disp_to_user_x(x),disp_to_user_y(y)); 
      statusline_echo(buf);
      if ((m.read(*this)).type() == MouseEvent::LBUTTONDOWN)
      {  m.get_position(x, y);
         double xx = disp_to_user_x(x);
         double yy = disp_to_user_y(y);
         Point p(xx, yy);
         statusline_clear();

         return p;
      }
   }
}

/*-------------------------------------------------------------------------*/

string GraphicWindow::get_string(string s)
{  char kbin;
   string s1;
   const int ENTER_KEY = 0x0D;
   const int BKSP_KEY = 0x08;

   statusline_prompt(s);

   statusline_echo("_");
   while((kbin = getch()) != ENTER_KEY)
   {  if(kbin == BKSP_KEY)
      {  if(s1.length() > 0)
            s1 = s1.substr(0, s1.length() - 1);
      }
      else
      {  s1 = s1 + kbin;
      }
      statusline_echo(s1 + "_");
   }
   // erase string input by user
   statusline_clear();
   return s1;
}

/*----------------------------------------------------------------------*/

static void find_bgi_dir(char* bgi_dir, int bgi_dir_size)
/* PURPOSE:    finds directory containing BGI files
   RECEIVES:   bgi_dir - system directory to be set up.
*/
{  char* path = getenv("PATH");
   char* path_end;
   char bgi_file[] = "\\..\\BGI\\EGAVGA.BGI";
   char bgi_path[] = "\\..\\BGI";
   do
   {  char try_file[100];
      path_end = strchr(path, ';');
      int length;
      if (path_end == NULL)
         length = strlen(path);
      else
         length = path_end - path;
      if (path[length - 1] == '\\') length--;
      if (length + sizeof(bgi_file) <= sizeof(try_file))
      {  strncpy(try_file, path, length);
         strcpy(try_file + length, bgi_file);
         if (access(try_file, 0) == 0
            && length + sizeof(bgi_path) <= bgi_dir_size)
         {  strncpy(bgi_dir, path, length);
            strcpy(bgi_dir + length, bgi_path);
            return;
         }
      }
      if (path_end == NULL) path = NULL;
      else path = path_end + 1;
   } while (path != NULL);
   bgi_dir[0] = '\0';
}

/*-------------------------------------------------------------------------*/

GraphicWindow::GraphicWindow()
:  _mouse(*this)
{  const double DEFAULT_XMIN = -10;
   const double DEFAULT_YMIN = 10;
   const double DEFAULT_XMAX = 10;
   const double DEFAULT_YMAX = -10;

   int gdriver = DETECT;
   int gmode;
   char bgi_dir[100];
   find_bgi_dir(bgi_dir, sizeof(bgi_dir));
   initgraph(&gdriver, &gmode, bgi_dir);
   int errorcode = graphresult();
   if (errorcode != grOk)
   {  cerr << "bgi_init: Can't initialize graphics. Is the .BGI file present?\n";
      exit(EXIT_FAILURE);
   }

   _disp_xoff = (getmaxx() - getmaxy()) / 2;
   _disp_yoff = 0;
   _disp_xmax = getmaxy() - _disp_yoff - 1;
   _disp_ymax = getmaxy() - _disp_yoff - 1;

   ::setviewport(_disp_xoff, _disp_yoff, getmaxx() - 1 - _disp_xoff, getmaxy() - 1, 1);

   ::setcolor(WHITE);
   ::setwritemode(COPY_PUT);
   coord(DEFAULT_XMIN, DEFAULT_YMIN, DEFAULT_XMAX, DEFAULT_YMAX);
}

//............................................................................

GraphicWindow::~GraphicWindow()
{  while (!bioskey(0))
    /* wait */ ;
   closegraph();
}

//............................................................................

void GraphicWindow::coord(double xmin, double ymin,
      double xmax, double ymax)
/* PURPOSE:    Set the logical coordinate system
   RECEIVES:   xmin, ymin - coordinates of the top left corner
               xmax, ymax - coordinates of the bottom right corner
*/
{  _user_xmin = xmin;
   _user_xmax = xmax;
   _user_ymin = ymin;
   _user_ymax = ymax;
}

//............................................................................

int GraphicWindow::user_to_disp_x(double x) const
/* PURPOSE:    scales the x coordinate from user to display coordinates
   RECEIVES:   x - a coordinate position in user coordinates
   RETURNS:    the scaled coordinate
*/
{  return (int)((x - _user_xmin) * (_disp_xmax - 2 * POINT_RADIUS) / (_user_xmax - _user_xmin)) + POINT_RADIUS;
}

//............................................................................

int GraphicWindow::user_to_disp_y(double y) const
/* PURPOSE:    scales the y coordinate from user to display coordinates
   RECEIVES:   y - a coordinate position in user coordinates
   RETURNS:    the scaled coordinate
*/
{  return (int)((y - _user_ymin) * (_disp_ymax - 2 * POINT_RADIUS) / (_user_ymax - _user_ymin)) + POINT_RADIUS;
}

//............................................................................

int GraphicWindow::user_to_disp_dx(double x) const
/* PURPOSE:    scales an x distance from user to display coordinates
   RECEIVES:   x - a distance in user coordinates
   RETURNS:    the scaled distance
*/
{  return (int)(x * (_disp_xmax - 2 * POINT_RADIUS) / (_user_xmax - _user_xmin));
}

//............................................................................

int GraphicWindow::user_to_disp_dy(double y) const
/* PURPOSE:    scales an y distance from user to display coordinates
   RECEIVES:   y - a distance in user coordinates
   RETURNS:    the scaled distance
*/
{  return (int)(y * (_disp_ymax  - 2 * POINT_RADIUS) / (_user_ymax - _user_ymin));
}

//............................................................................

double GraphicWindow::disp_to_user_dx(int x) const
/* PURPOSE:    scales an x distance from display to user coordinates
   RETURNS:    the scaled distance
*/
{  return (double)x * (_user_xmax - _user_xmin) / (_disp_xmax  - 2 * POINT_RADIUS);
}

//............................................................................

double GraphicWindow::disp_to_user_dy(int y) const
/* PURPOSE:    scales an y distance from user to display coordinates
   RETURNS:    the scaled distance
*/
{  return (double)y * (_user_ymax - _user_ymin) / (_disp_ymax - 2 * POINT_RADIUS);
}

//............................................................................

double GraphicWindow::disp_to_user_x(int x) const
/* PURPOSE:    scales an x distance from display to user coordinates
   RETURNS:    the scaled distance
*/
{  return (double)(x - _disp_xoff - POINT_RADIUS) * (_user_xmax - _user_xmin) / (_disp_xmax - 2 * POINT_RADIUS) + _user_xmin;
}

//............................................................................

double GraphicWindow::disp_to_user_y(int y) const
/* PURPOSE:    scales an y distance from user to display coordinates
   RETURNS:    the scaled distance
*/
{  return (double)(y - _disp_yoff - POINT_RADIUS) * (_user_ymax - _user_ymin) / (_disp_ymax - 2 * POINT_RADIUS) + _user_ymin;
}

//............................................................................

void GraphicWindow::point(double x, double y)
/* PURPOSE:    plot a point
   RECEIVES:   x, y - the point coordinates
*/
{  int cx = user_to_disp_x(x);
   int cy = user_to_disp_y(y);
   ::ellipse(cx, cy, 0, 360, POINT_RADIUS, POINT_RADIUS);
}

//............................................................................

void GraphicWindow::line(double xfrom, double yfrom, double xto, double yto)
/* PURPOSE:    plot a line segment
   RECEIVES:   xfrom, yfrom - starting point
               xto, yto - ending point
*/
{  int x1 = user_to_disp_x(xfrom);
   int x2 = user_to_disp_x(xto);
   int y1 = user_to_disp_y(yfrom);
   int y2 = user_to_disp_y(yto);
   ::line(x1, y1, x2, y2);
}

//............................................................................

void GraphicWindow::clear()
/* PURPOSE:    Clear the display
*/
{  ::clearviewport();
}

//............................................................................

void GraphicWindow::ellipse(double x, double y, double ra, double rb)
/* PURPOSE:    Plot an ellipse
   RECEIVES:   x, y - center
               ra, rb - radii in x- and y- direction
   REMARKS:    We are breaking the ellipse into a polygon since
               the XOR rop mode is not "currently" supported for ellipses
*/

{  int cx = user_to_disp_x(x);
   int cy = user_to_disp_y(y);
   int a = abs(user_to_disp_dx(ra));
   int b = abs(user_to_disp_dy(rb));
   ::ellipse(cx, cy, 0, 360, a, b);
}

//............................................................................

void GraphicWindow::text(string t, double x, double y)
/* PURPOSE:    plot text in the current font
   RECEIVES:   t - the text
               x, y - the top left corner
*/
{  int dx = user_to_disp_x(x);
   int dy = user_to_disp_y(y);
   outtextxy(dx, dy, t.c_str());
}

//............................................................................

GraphicWindow& GraphicWindow::operator<<(Line s)
{  line(s.get_start().get_x(), s.get_start().get_y(),
         s.get_end().get_x(), s.get_end().get_y());
   return *this;
}

//............................................................................

GraphicWindow& GraphicWindow::operator<<(Message t)
{
   text(t.get_message(), t.get_start().get_x(), t.get_start().get_y());
   return *this;
}

//............................................................................//............................................................................//............................................................................

GraphicWindow& GraphicWindow::operator<<(Circle c)
{
   ellipse(c.get_center().get_x(), c.get_center().get_y(),
         c.get_radius(), c.get_radius());
   return *this;
}

//............................................................................//............................................................................//............................................................................

GraphicWindow& GraphicWindow::operator<<(Point p)
{  point (p.get_x(), p.get_y());
   return *this;
}

//............................................................................//............................................................................//............................................................................

void GraphicWindow::statusline_prompt(string s)
{  _prompt = s;
   outtextxy(0, 0, _prompt.c_str());
}

//............................................................................//............................................................................//............................................................................

void GraphicWindow::statusline_echo(string s)
{  ::setcolor(BLACK);
   string p = _prompt + " ";
   int w = textwidth(p.c_str());
   outtextxy(w, 0, _echo.c_str());
   ::setcolor(WHITE);
   _echo = s;
   outtextxy(w, 0, _echo.c_str());
}

//............................................................................//............................................................................//............................................................................

void GraphicWindow::statusline_clear()
{  ::setcolor(BLACK);
   outtextxy(0, 0, _prompt.c_str());
   string p = _prompt + " ";
   int w = textwidth(p.c_str());
   outtextxy(w, 0, _echo.c_str());
   _prompt = "";
   _echo = "";
   ::setcolor(WHITE);
}

/*-------------------------------------------------------------------------*/

MouseEvent::MouseEvent(Type t, double x, double y)
/* RECEIVES:   t - event type
               x, y - mouse coordinates
*/
:  _type(t), _x(x), _y(y)
{}

/*-------------------------------------------------------------------------*/

Mouse::Mouse(GraphicWindow& w)
{  unsigned char* mouse_vect = (unsigned char*)_dos_getvect(MOUSE_INTR);
   _installed = mouse_vect != 0 && *mouse_vect != RETF;
   ws = &w;
}


//............................................................................

bool Mouse::is_installed() const
/* RETURNS:    true iff a mouse is installed
*/
{  return _installed;
}

//............................................................................

void Mouse::show_cursor()
/* PURPOSE:    show mouse cursor
*/
{  if (!_installed) return;
   union REGS regs;
   regs.x.ax = SHOWCURSOR;
   int86(MOUSE_INTR, &regs, &regs);
}

//............................................................................

void Mouse::hide_cursor()
/* PURPOSE:    hide mouse cursor
*/
{  if (!_installed) return;
   union REGS regs;
   regs.x.ax = HIDECURSOR;
   int86(MOUSE_INTR, &regs, &regs);
}

//............................................................................

int Mouse::get_position(int& x, int& y) const
/* PURPOSE:    get current mouse position
   RECEIVES:   x, y (OUT) - set to mouse position
   RETURNS:    mouse button flags (bit #0 = left button, #1 = middle button,
          #2 = right button)
*/
{  if (!_installed) return 0;
   union REGS regs;
   regs.x.ax = READMOUSE;
   int86(MOUSE_INTR, &regs, &regs);
   x = regs.x.cx;
   y = regs.x.dx;
   return regs.x.bx;
}

//............................................................................

MouseEvent Mouse::read(const GraphicWindow& gc)
/* PURPOSE:    get next mouse event
   RECEIVES:   gc - graphics context (to convert mouse coordinates to graphics)
   RETURNS:    the next mouse event
*/
{  if (!_installed) return MouseEvent(MouseEvent::NOT_INSTALLED, 0, 0);
   // get buttons
   // get x, y
   int old_x, old_y, old_b;
   int x, y, b;
   show_cursor();
   old_b = get_position(old_x, old_y) & 1;
   bool changed = false;
   while (!changed) // buttons and x, y not changed
   {  b = get_position(x, y) & 1;
      if (old_x != x || old_y != y || old_b != b) changed = true;

   }
   MouseEvent::Type t;
   if (b == old_b) t = MouseEvent::MOUSEMOVE;
   else if (b) t = MouseEvent::LBUTTONDOWN;
   else t = MouseEvent::LBUTTONUP;
   hide_cursor();

   double dx = gc.disp_to_user_x(x);
   double dy = gc.disp_to_user_y(y);

   return MouseEvent(t, dx, dy);
}

/***************************************************************************/

#define main ccc_win_main
