/****************************************************************************
** COPYRIGHT (C):    1998 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Computing Concepts with C++ 2E
** FILE:             ccc_list.cpp
** PURPOSE:          A subset of the ANSI list template 
** VERSION           1.0
** PROGRAMMERS:      Cay S. Horstmann (CSH)
****************************************************************************/

/* IMPORT ******************************************************************/

#define EXPORT

#ifndef EXPORT
   #ifdef CCC_USE_OLD_CLIBS
      #include <stdlib.h>
      #include <assert.h>
   #else
      #include <cstdlib>
      #include <cassert>
   #endif
#endif

#include "ccc_list.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

template<class T>
class list
{
private:

   class link
   {
   public:
      link* previous;
      link* next;
      T data;

      link(T s)
      /* RECEIVES:  s - the data to store in this link
      */
      {  data = s;
         previous = NULL;
         next = NULL;
      }

   };


   link* first;
   link* last;

public:

   class iterator
   {
   public:
      iterator()
      {  position = NULL;
         last = NULL;
      }

      T operator*() const
      /* PURPOSE:  look up the value at a position
         RETURNS:  the value of the Link<T> to which the iterator points
      */
      {  assert(position != NULL);
         return position->data;
      }

      void operator++(int dummy)
      /* PURPOSE:  advance the iterator to the next Link<T>
      */
      {  assert(position != NULL);
         position = position->next;
      }

      void operator--(int dummy)
      /* PURPOSE:  move the iterator to the previous Link<T>
      */
      {  if (position == NULL)
            position = last;
         else 
            position = position->previous;
         assert(position != NULL);
      }

      bool operator==(iterator b) const
      /* PURPOSE:  compare two iterators
         RECEIVES: b - the iterator to compare with this iterator
         RETURNS:  true if this iterator and b are not equal
      */
      {  return position == b.position;
      }

      bool operator!=(iterator b) const
      /* PURPOSE:  compare two iterators
         RECEIVES: b - the iterator to compare with this iterator
         RETURNS:  true if this iterator and b are not equal
      */
      {  return position != b.position;
      }

   private:
      link* position;
      link* last;
   friend class list<T>;
   };


   list()
   {  first = NULL;
      last = NULL;
   }

   ~list()
   {  free();
   }

   list(const list<T>& b)
   {  copy(b);
   }

   list<T>& operator=(const list<T>& b)
   {  if (this != &b)
      {  free();
         copy(b);
      }
      return *this;
   }

   void push_back(T s)
   /* PURPOSE:  append an element to the list
      RECEIVES: s - the value to append
   */
   {  link* newlink = new link(s);
      if (last == NULL) /* list is empty */
      {  first = newlink;
         last = newlink;
      }
      else
      {  newlink->previous = last;
         last->next = newlink;
         last = newlink;
      }
   }

   void insert(iterator iter, T s)
   /* PURPOSE:  insert an element into the list
      RECEIVES: iter - the position before which to insert
                s - the value to append
   */
   {  if (iter.position == NULL)
      {  push_back(s);
         return;
      }

      link* after = iter.position;
      link* before = after->previous;
      link* newlink = new link(s);
      newlink->previous = before;
      newlink->next = after;
      after->previous = newlink;
      if (before == NULL) /* insert at beginning */
         first = newlink;
      else
         before->next = newlink;
   }

   void erase(iterator iter)
   /* PURPOSE:  remove an element from the list
      RECEIVES: iter - the position to remove
   */
   {  assert(iter.position != NULL);
      link* remove = iter.position;
      link* before = remove->previous;
      link* after = remove->next;
      if (remove == first)
         first = after;
      else
         before->next = after;
      if (remove == last)
         last = before;
      else
         after->previous = before;
      iter.position = after;
      delete remove;
   }

   iterator begin() const
   /* PURPOSE:  get the beginning position of the list
      RETURNS:  an iterator pointing to the beginning of the list
   */
   {  iterator iter;
      iter.position = first;
      iter.last = last;
      return iter;
   }

   iterator end() const
   /* PURPOSE:  get the past-the-end position of the list
      RETURNS:  an iterator pointing past the end of the list
   */
   {  iterator iter;
      iter.position = NULL;
      iter.last = last;
      return iter;
   }

private:
   void copy(const list<T>& b);
   void free();
};

template<class T>
void list<T>::copy(const list<T>& b)
{  first = NULL;
   last = NULL;
   iterator p;
   for (p = b.begin(); p != b.end(); p++)
      push_back(*p);
}

template<class T>
void list<T>::free()
{  while (begin() != end())
      erase(begin());
}

#endif