/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
**
** NOTE TO STUDENTS: Do not attempt to study the contents of this file. You
** can, and should, use the services provided in this file without knowing
** the highly technical details of the implementation.
**
** PROJECT:          Computing Concepts with C++ Essentials
** FILE:             ccc_x11.cpp
** PURPOSE:          Graphics context class (X-Window's version)
** VERSION           2.0
** COMPILER:         GNU g++
** TARGET:           486, Linux
** PROGRAMMER:       Cay Horstmann (CSH)
**                   Darrel D. Cherry (DDC)
**                   Charles Miller (CBM)
**                   Geof Pawlicki (GFP)
** START DATE:       6-5-94  (CBM)
** UPDATE HISTORY:   8-2-95  (DDC)
**                   9-25-95 (DDC) - fixed repainting in read string/point
**                   6-1-97  (CSH) - removed font and colors
****************************************************************************/

/* IMPORT ******************************************************************/

#define EXPORT

#ifndef EXPORT

#ifdef CCC_USE_OLD_CLIBS
   #include <math.h>
   #include <stdlib.h>
#else
   #include <cmath>
   #include <cstdlib>
#endif

#define Time X11_TIME    // so that Time class doesn't conflict with #typedef
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/Xos.h>
#undef Time

#endif
// EXPORT


extern char* ltoa (long, char* , int);

#include "ccc_x11.h"

/* CONSTANTS ****************************************************************/

const double DEFAULT_XMIN = -10;
const double DEFAULT_YMIN = 10;
const double DEFAULT_XMAX = 10;
const double DEFAULT_YMAX = -10;

static int DEF_WIDTH = 480;
static int DEF_HEIGHT = 480;

/* EXPORT ******************************************************************/

/* Our function wrapper. */
int ccc_win_main();

// The following are some aliases for various X fonts.
// Make sure these are available on your system.  Substitute as necessary.
#define SYS_14  "-adobe-courier-bold-r-normal-*-14-*-*-*-*-*-*"
#define SYS_FONT_NAME SYS_14

/* CLASSES *****************************************************************/

#ifndef EXPORT

class GraphicBitmap
/* PURPOSE: Provides pixmap memory for screen redraws
   NOTES:  Make sure global xgc has been initialized.
*/
{
public:
   GraphicBitmap();
   void init();
   void close();
   void repaint() const;
   void clear_bitmap(unsigned long color) const;
   Pixmap id() const;
private:
   Pixmap _ppm;
};


/**************************************************************************/

class GraphicWindow
/* PURPOSE: Sends graphics primitives to a display window
*/
{
public:
   GraphicWindow();
   GraphicWindow(double x1, double y1, double x2, double y2);
   void coord(double xmin, double ymin, double xmax, double ymax);
   void close();
   void clear() const;
   void common_init();

   /*** overloaded operators ***/

   GraphicWindow& operator<<(Point p);
   GraphicWindow& operator<<(Circle c);
   GraphicWindow& operator<<(Line l);
   GraphicWindow& operator<<(Message m);

   /*** User prompted input  ***/

   string get_string(string s);
   int get_int(const string& out_string) 
   {  return atoi(get_string(out_string).c_str()); 
   }
   double get_double(const string& out_string) 
   {  return atof(get_string(out_string).c_str()); 
   }
   Point get_mouse(string outstr);

private:

   /***  I/O  ***/     

   int user_to_disp_x(double x) const;
   int user_to_disp_y(double y) const;
   int user_to_disp_dx(double x) const;
   int user_to_disp_dy(double y) const;
   double disp_to_user_x(int x) const;
   double disp_to_user_y(int y) const;

   void statusline_prompt(string s);

   /***  Shapes  ***/     

   void point(double x, double y) const;
   void line(double xfrom, double yfrom, double xto, double yto) const;
   void ellipse(double x, double y, double ra, double rb) const;
   void text(const char t[], double x, double y) const;

   int put_string(const char[], int, int);

   string _display_string;    // string displayed to user at top

   double _user_xmin;            // the user's window's logical window coords
   double _user_ymin;            // min is upper left   
   double _user_xmax;            // max lower right 
   double _user_ymax;
   int _disp_xmax;               // the physical window dimension (in pixels)
   int _disp_ymax;

   XFontStruct* _fontinfo_ptr;    // points to font info for all fonts used
};

#endif
  // EXPORT

/* GLOBALS *************************************************************/

EXPORT GraphicWindow cwin;

Window win; 

Display* display;
int screen_num;
GC xgc;
GraphicBitmap mainwin_bm;

/* OPERATIONS **************************************************************/

GraphicBitmap::GraphicBitmap()
/* PURPOSE: Default constructor.  We defer data initialization.
*/
{}

//..........................................................................

void GraphicBitmap::init()
/* PURPOSE: Allocates an X pixmap
*/
{  _ppm = XCreatePixmap(display, win, DEF_WIDTH, DEF_HEIGHT, DefaultDepth(display, screen_num));
}

//..........................................................................

void GraphicBitmap::close() 
/* PURPOSE: Returns pixmap memory to the freestore.
*/
{  XFreePixmap(display, _ppm);
}

//..........................................................................

void GraphicBitmap::repaint() const
/* PURPOSE: Puts the image of the pixmap onto the display screen.
*/
{  XCopyArea(display, _ppm, win, xgc, 0, 0, DEF_WIDTH, DEF_HEIGHT, 0, 0);
}

//..........................................................................

void GraphicBitmap::clear_bitmap(unsigned long color) const
/* PURPOSE: Clears the pixmap, clearing with color color.
*/
{  XGCValues vals;
   XGetGCValues(display, xgc, GCForeground, &vals);

   unsigned long fg = vals.foreground;
   XSetForeground(display, xgc, color);
   XFillRectangle(display, _ppm, xgc, 0, 0, DEF_WIDTH + 1, DEF_HEIGHT + 1);
   XSetForeground(display, xgc, fg);
}

//..........................................................................

Pixmap GraphicBitmap::id() const
/* PURPOSE: Returns the pixmap id.
*/
{  return _ppm;
}

/**************************************************************************/

GraphicWindow::GraphicWindow()
/* PURPOSE: Initializes user coordinates
   RECEIVES: The logical coordinates for your window
   NOTES: Initializes both GC and pixmap drawing tools
*/
:  _user_xmin(DEFAULT_XMIN), _user_ymin(DEFAULT_YMIN), 
   _user_xmax(DEFAULT_XMAX), _user_ymax(DEFAULT_YMAX),
   _disp_xmax(DEF_WIDTH), _disp_ymax(DEF_HEIGHT)
{
}

//..........................................................................

GraphicWindow::GraphicWindow(double x1, double y1, double x2, double y2)
/* PURPOSE: Initializes user coordinates
   RECEIVES: The logical coordinates for your window
   NOTES: Initializes both GC and pixmap drawing tools
*/
:  _user_xmin(x1), _user_ymin(y1), _user_xmax(x2), _user_ymax(y2),
   _disp_xmax(DEF_WIDTH), _disp_ymax(DEF_HEIGHT)
{
}

//...........................................................................

void GraphicWindow::common_init()
/* PURPOSE:  common code for multiple constructors
*/
{  //   specify black foreground since default window background is
   //   white and default foreground is undefined.
   XSetForeground(display, xgc, BlackPixel(display, screen_num));

   // set line attributes.  This class only uses rounded styles.
   XSetLineAttributes(display, xgc, 2 /* pen thickness */, LineSolid,
         CapRound, JoinRound);

   // set up the font - uses only 14pt bold Courier
   _fontinfo_ptr = XLoadQueryFont(display, SYS_FONT_NAME);
   if (_fontinfo_ptr == NULL)
   {   cerr << "Unable to initialize font: " << SYS_FONT_NAME << endl;
      exit(1);
    }
   XSetFont(display, xgc, _fontinfo_ptr->fid);

   // sets the rop mode, using Copy for the user prompts
   XSetFunction(display, xgc, GXcopy /*0x3 COPY*/);

   clear();
}

//...........................................................................

void GraphicWindow::close()
/* PURPOSE:    Deallocates various objects when gc is through.
   NOTES:      This does NOT close down the private pixmap--only deallocates
               the tools this gc uses.
*/
{
    XFreeFont(display, _fontinfo_ptr);
}

//...........................................................................

void GraphicWindow::clear() const
/* PURPOSE:    Clears the DC the gc is in, and the pixmap too.
*/
{
   XClearWindow(display, win);
   mainwin_bm.clear_bitmap(WhitePixel(display, screen_num));
}

//...........................................................................

void GraphicWindow::coord(double xmin, double ymin,
      double xmax, double ymax)
/* PURPOSE:    Set the logical coordinate system
   RECEIVES:   xmin, ymin - coordinates of the top left corner
               xmax, ymax - coordinates of the bottom right corner
*/
{  _user_xmin = xmin;
   _user_xmax = xmax;
   _user_ymin = ymin;
   _user_ymax = ymax;
}

//...........................................................................

int GraphicWindow::user_to_disp_x(double x) const
/* PURPOSE:    scales the x coordinate from user to display coordinates
   RECEIVES:   x - a coordinate position in user coordinates
   RETURNS:    the scaled coordinate
*/
{  return (int) ((x - _user_xmin) * _disp_xmax / (_user_xmax - _user_xmin));
}

//...........................................................................

int GraphicWindow::user_to_disp_y(double y) const
/* PURPOSE:    scales the y coordinate from user to display coordinates
   RECEIVES:   y - a coordinate position in user coordinates
   RETURNS:    the scaled coordinate
*/
{  return (int) ((y - _user_ymin) * _disp_ymax / (_user_ymax - _user_ymin)); 
}

//...........................................................................

int GraphicWindow::user_to_disp_dx(double x) const
/* PURPOSE:    scales an x distance from user to display coordinates
   RECEIVES:   x - a distance in user coordinates
   RETURNS:    the scaled distance
*/
{  return (int) (x * _disp_xmax / (_user_xmax - _user_xmin));
}

//...........................................................................

int GraphicWindow::user_to_disp_dy(double y) const
/* PURPOSE:    scales an y distance from user to display coordinates
   RECEIVES:   y - a distance in user coordinates
   RETURNS:    the scaled distance
*/
{  return (int) (y * _disp_ymax / (_user_ymax - _user_ymin));
}


//..........................................................................

double GraphicWindow::disp_to_user_x(int x) const
/* PURPOSE:    scales an x distance from display to user coordinates
   RETURNS:    the scaled distance
*/

{  return (double)x * (_user_xmax - _user_xmin) / _disp_xmax + _user_xmin;
}

//...........................................................................

double GraphicWindow::disp_to_user_y(int y) const
/* PURPOSE:    scales an y distance from display to user coordinates
   RETURNS:    the scaled distance
*/

{  return (double)y * (_user_ymax - _user_ymin) / _disp_ymax + _user_ymin;
}

//..........................................................................

void GraphicWindow::point(double x, double y) const
/* PURPOSE:    plot a point
   RECEIVES:   x, y - the point coordinates
*/
{  const int POINT_RADIUS = 2;
   int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);

   XSetForeground(display, xgc, BlackPixel(display, screen_num));
   XDrawArc(display, win, xgc, disp_x - POINT_RADIUS, disp_y - POINT_RADIUS,
      POINT_RADIUS, POINT_RADIUS, 0 , 360 * 64);
   XDrawArc(display, mainwin_bm.id(), xgc, disp_x - POINT_RADIUS,
      disp_y - POINT_RADIUS, POINT_RADIUS, POINT_RADIUS, 0 , 360 * 64);
}

//...........................................................................

void GraphicWindow::ellipse(double x, double y, double ra, double rb) const
/* PURPOSE:    Plot an ellipse
   RECEIVES:   x, y - center
               ra, rb - radii in x- and y- direction
*/
{  int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);
   int disp_rx = abs(user_to_disp_dx(ra));
   int disp_ry = abs(user_to_disp_dy(rb));

   XSetForeground(display, xgc, BlackPixel(display, screen_num));
   XDrawArc(display, win, 
      xgc,               // the drawable
      disp_x - disp_rx,  // the x coord of upperleft corner of bounding rect
      disp_y - disp_ry,  // the y coord of upperleft corner of bounding rect
      disp_rx * 2,       // the major axis length
      disp_ry * 2,       // the minor axis length
      0,                 // offset from 3 o'clock  
      360 * 64);         // complete circle (360 degrees * 64 clicks per degree)
   XDrawArc(display, mainwin_bm.id(), xgc, disp_x - disp_rx, 
   disp_y - disp_ry, disp_rx * 2, disp_ry * 2, 0, 360 * 64);
}

//...........................................................................

void GraphicWindow::line(double xfrom, double yfrom, double xto,
   double yto) const
/* PURPOSE:    plot a line segment with current pen style and color
   RECEIVES:   xfrom, yfrom - starting point
               xto, yto - ending point
*/
{  XSetForeground(display, xgc, BlackPixel(display, screen_num));
   XDrawLine(display, win, xgc, user_to_disp_x(xfrom), user_to_disp_y(yfrom),
      user_to_disp_x(xto), user_to_disp_y(yto));
   XDrawLine(display, mainwin_bm.id(), xgc, user_to_disp_x(xfrom),
      user_to_disp_y(yfrom), user_to_disp_x(xto), user_to_disp_y(yto));
}

//...........................................................................

void GraphicWindow::text(const char t[], double x, double y) const
/* PURPOSE:    plot text in the current font
   RECEIVES:   t - the text
               x, y - coordinates to start plotting
*/
{  int disp_x = user_to_disp_x(x);
   int disp_y = user_to_disp_y(y);

   XSetForeground(display, xgc, BlackPixel(display, screen_num));
   XDrawString(display, win, xgc, disp_x, disp_y, t, strlen(t));
   XDrawString(display, mainwin_bm.id(), xgc, disp_x, disp_y, 
      t, strlen(t));
}

//...........................................................................

GraphicWindow& GraphicWindow::operator<<(Point p)
/* PURPOSE: outputs a point to the main window
   RECEIVES: the point
*/
{  point(p.get_x(), p.get_y());
   return *this;
}

//...........................................................................

GraphicWindow& GraphicWindow::operator<<(Circle c)
/* PURPOSE: output a circle to the main window
   RECEIVES: the circle
*/

{  ellipse (c.get_center().get_x(), 
      c.get_center().get_y(),
      c.get_radius(),
      c.get_radius());
   return *this;
}


//...........................................................................

GraphicWindow& GraphicWindow::operator<<(Message m)
/* PURPOSE: output a Message to the main window
   RECEIVES: the message
*/
{  _display_string = m.get_message();
   text(_display_string.c_str(), m.get_start().get_x(), m.get_start().get_y());

   return *this;
}

//...........................................................................

GraphicWindow& GraphicWindow::operator<< (Line l)
/* PURPOSE: output a line segment to the main window
   RECEIVES: the line segment
*/
{  line(l.get_start().get_x(), l.get_start().get_y(), 
       l.get_end().get_x(), l.get_end().get_y());
   return *this;
}


//...........................................................................

void GraphicWindow::statusline_prompt(string s)
/* PURPOSE: outputs a string (containing status message or prompting for input)
   RECEIVES: the string
*/
         
{  int direction, ascent, descent;
   XCharStruct overall;
   const char* t = s.c_str();
   XTextExtents(_fontinfo_ptr, t, strlen(t), 
      &direction, &ascent, &descent, &overall);
  
   XSetBackground(display, xgc, WhitePixel(display, screen_num));

   XClearArea(display, win, 
                0,
                0, /*clear from top*/
                0 /*clear to right edge of screen*/, 
                ascent + descent, 
                0 /* generate exposure events */);

   XDrawImageString(display, win, xgc, 0, ascent, t, strlen(t));
   XDrawImageString(display, mainwin_bm.id(), xgc, 0, ascent, t, strlen(t));    
}


/***************************************************************************/

Point GraphicWindow::get_mouse(string outstr)
/* PURPOSE: read the window point that the user clicked with the mouse
   RECEIVES: string prompting for a click
   RETURNS: the point
*/
{  XEvent report;
   Point p;
   int x, y;

   cwin.statusline_prompt(outstr);   

   // Event Loop
   while (1)
   {
      XNextEvent(display, &report);
      switch(report.type)
      {
         case Expose:
             mainwin_bm.repaint();
             break;

         case ButtonPress:
            if (report.xbutton.button == Button1)
            {
               x = report.xbutton.x;
               y = report.xbutton.y;
               p = Point(cwin.disp_to_user_x(x), cwin.disp_to_user_y(y));
               cwin.statusline_prompt(outstr);
               return p;               
            }
       }
   }

   return p; // to keep compiler happy, will never get here!
}
                

/***************************************************************************/

int GraphicWindow::put_string(const char instr[], int str_x, int curr_x)
/* PURPOSE: 
   RECEIVES:
   RETURNS:
*/
{  int direction, ascent, descent;
   XCharStruct overall;
   XTextExtents(_fontinfo_ptr, instr, strlen(instr), &direction, &ascent, &descent, &overall);

   XClearArea(display, win, 
              str_x - 1,
              0, /*clear from top*/
              curr_x + 1, /*clear to right edge of screen*/ 
              ascent + descent + 1, 
              0 /* generate exposure events */);

   XSetForeground(display, xgc, BlackPixel(display, screen_num));
   XDrawImageString(display, win, xgc, str_x, ascent, instr, strlen(instr));
   curr_x = str_x + overall.width;
   /* redraw cursor */
   XDrawLine(display, win, xgc, curr_x, 0, curr_x, ascent + descent);
   return curr_x;
}

/***************************************************************************/

string GraphicWindow::get_string(string outstr)
/* PURPOSE: read text from the edit window
   RECEIVES: out_string - text to prompt the user
   RETURNS: the string read in from the edit window
*/
{  XEvent report;
   string instring;
   char instr[50];
   char buffer[3];
   int   buffsize = 3;
   KeySym keysym;
   XComposeStatus compose;
   int count, length;

   int str_x;        // initial x position, at the end of the prompt string
   int str_y;        // initial yposition
   int curr_x;       // current x position (measured from str_x)

   int direction;    // draw chars l->r
   int ascent;       // above font's baseline
   int descent;      // below font's baseline
   XCharStruct overall;

   _display_string=outstr;

   statusline_prompt(outstr);

   XTextExtents(_fontinfo_ptr, _display_string.c_str(), 
      _display_string.length(), &direction, &ascent, &descent, &overall);

   str_x = overall.width; 
   curr_x = overall.width;
   str_y = ascent;   // set initial cursor position 

   instr[0] = '\0';
   curr_x = put_string(instr, str_x, curr_x);

   //Event loop
   while(1)
   {
      XNextEvent(display, &report);
      switch (report.type)
      {
         case Expose:
             mainwin_bm.repaint();
             break;

         case KeyPress:

            /* get characters until carriage return */
            count = XLookupString(&(report.xkey), buffer, buffsize, &keysym, &compose);

            /* got a carriage return */
            if ((keysym == XK_Return) || (keysym == XK_KP_Enter)
                || (keysym == XK_Linefeed))
            {
                 XClearArea(display, win, 
                            0,
                            0, /*clear from top*/
                            0, /*clear to right edge of screen*/ 
                            ascent + descent + 1, 
                            0 /* generate expoure events */);

                 instring = instr;
                 return instring;
            }

            /* eat white space */
            else if (((keysym >= XK_KP_Space) && (keysym <= XK_KP_9))
                     || ((keysym >= XK_space) && (keysym <= XK_asciitilde)))
            {
               if ((strlen(instr) + strlen(buffer)) >= 50)
                  XBell(display, 100);
               else
               {
                  buffer[1] = '\0';
                  strcat (instr, buffer);
               }
            }

            /* got a delete key */
            else if ((keysym == XK_BackSpace) || (keysym == XK_Delete))
            {
                if ((length = strlen(instr)) > 0)
                {  
                   // resize the string 
                   instr[length-1] = '\0';
                }
                else 
                    XBell(display, 100);
            }            

            /* display the new string, and reposition cursor */     
            curr_x = put_string(instr, str_x, curr_x);
      }
   }
   return instring; // never get here but it keeps compiler happy :)!
}

/***************************************************************************/

void close(GraphicWindow& cwin)
{  cwin.close();
}

/***************************************************************************/

#define BITMAPDEPTH 1
#define TOO_SMALL 0
#define BIG_ENOUGH 1

/***************************************************************************/

int main(int argc, char** argv)
{  // First, X overhead...
   unsigned int width, height;   // window size
   unsigned int border_width = 4;

   char* window_name = argv[0];
   char* icon_name = "GCT";

   Pixmap icon_pixmap;
   XSizeHints* size_hints;
   XWMHints* wm_hints;
   XClassHint* class_hints;
   XTextProperty windowName;
   XTextProperty iconName;
   XEvent report;
   int window_size_ok = 1;
   char* display_name = 0;
   char* progname = argv[0];

   // check allocation of hints
   if (  !(size_hints = XAllocSizeHints())
      || !(wm_hints = XAllocWMHints())
      || !(class_hints = XAllocClassHint()))
   {  cerr << progname << " error: failure allocating memory" << endl;
      exit(0);
   }

   progname = argv[0];

   // connect to X server
   display = XOpenDisplay(display_name);
   if (display == NULL)
   {  cerr << progname << " error: can't connect to server " <<
         XDisplayName(display_name) << endl;
      exit(-1);
   }

   // get screen size from display structure macro
   screen_num = DefaultScreen(display);

   win = XCreateSimpleWindow(display, RootWindow(display, screen_num),
      0, 0, DEF_WIDTH, DEF_HEIGHT, border_width, BlackPixel(display,
      screen_num), WhitePixel(display, screen_num));

   if (XStringListToTextProperty(&window_name, 1, &windowName) == 0)
   {  cerr << progname
         << " error: structure allocation for windowName failed."
         << endl;
      exit(-1);
   }

   if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) 
   {  cerr << progname
         << " error: structure allocation for iconName failed."
         << endl;
      exit(-1);
   }

   size_hints->flags = PPosition | PSize | PMinSize;
   size_hints->min_width = DEF_WIDTH;
   size_hints->min_height = DEF_HEIGHT;

   wm_hints->initial_state = NormalState;
   wm_hints->input = True;
   wm_hints->icon_pixmap = icon_pixmap;
   wm_hints->flags = StateHint | IconPixmapHint | InputHint;

   class_hints->res_name = progname;
   class_hints->res_class = "chigcx";

   XSetWMProperties(display, win, &windowName, &iconName, argv, argc,
      size_hints, wm_hints, class_hints);

   // Select desired event types
   XSelectInput(display, win, ExposureMask | KeyPressMask
      | ButtonPressMask | StructureNotifyMask);

   XGCValues xgc_values;  // ignored
   xgc = XCreateGC(display, win, 0, &xgc_values);
   XMapWindow(display, win);

   mainwin_bm.init();
   cwin.common_init();

   // Event loop
   static int draw_flag = 1;
   while (1)
   {  XNextEvent(display, &report);
      switch ((int)report.type)
      {  case Expose:
            if (report.xexpose.count == 0)
            {  if (window_size_ok)
               {  if (draw_flag)
                  {  ccc_win_main();
                     draw_flag = 0;
                  }
                  else
                     mainwin_bm.repaint();
               }
            }
            break;

         case ConfigureNotify:
            /* window has been resized, change width and
             * height to send to draw_text and draw_graphics
             * in next Expose */
            width = report.xconfigure.width;
            height = report.xconfigure.height;
            window_size_ok = (width >= size_hints->min_width)
               && (height >= size_hints->min_height);
            break;

         case ButtonPress:
         case KeyPress:
         default:
            break;
      }
   }
   return 0;
}

/***************************************************************************/

#define main ccc_win_main

