/****************************************************************************
** COPYRIGHT (C):    1996 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Computing Concepts with C++
** FILE:             clock.cpp
****************************************************************************/

#include "ccc_win.cpp"
#include "ccc_time.cpp"

const double PI = 3.141592653589793;

class Clock
{
public:
   Clock();
   Clock(Point c, double r);

   void set_time(Time t);

   void draw() const;
private:
   void draw_tick(double angle, double length) const;
   void draw_hand(double angle, double length) const;

   Time current_time;
   Point center;
   double radius;
};

class Player
{
public:
   Player();
   Player(string player_name, int initial_level);

   void increment_score();

   int get_level() const;
   string get_name() const;
private:
   string name;
   int score;
   int level;
};

class Game
{
public:
   Game();

   void play();
   void read_player_information();
   void play_round();
   Time random_time();
   Time get_guess();
private:
   Player player;
};

void rand_seed()
/* PURPOSE: Set the seed of the random number generator
*/
{  int seed = static_cast<int>(time(0));
   srand(seed);
}

int rand_int(int a, int b)
/* PURPOSE:  Return a random integer in a range
   RECEIVES: a - the bottom of the range
             b - the top of the range
   RETURNS:  A random number x, a <= x and x <= b
*/
{  return a + rand() % (b - a + 1);
}

Clock::Clock()
{
}

Clock::Clock(Point c, double r)
/* RECEIVES:  c - the center of the clock
              r - the radius of the clock
*/              
{  center = c;
   radius = r;
}

void Clock::set_time(Time t)
/* PURPOSE:  set the current time
   RECEIVES: t - the time to set
*/
{  current_time = t;
}

void Clock::draw_tick(double angle, double length) const
/* PURPOSE:  draw a tick mark (hour or minute mark)
   RECEIVES: angle - the angle in minutes (0...59, 0 = top)
             length - the length of the tick mark
*/
{  double alpha = PI / 2 - 6 * angle * PI / 180;
   Point from(center.get_x() + cos(alpha) * radius * (1 - length),
      center.get_y() + sin(alpha) * radius * (1 - length));
   Point to(center.get_x() + cos(alpha) * radius,
      center.get_y() + sin(alpha) * radius);
   cwin << Line(from, to);
}

void Clock::draw_hand(double angle, double length) const
/* PURPOSE:  draw a hand, starting from the center
   RECEIVES: angle - the angle in minutes (0...59, 0 = top)
             length - the length of the hand
*/
{  double alpha = PI / 2 - 6 * angle * PI / 180;
   Point from = center;
   Point to(center.get_x() + cos(alpha) * radius * length,
      center.get_y() + sin(alpha) * radius * length);
   cwin << Line(from, to);
}

void Clock::draw() const
/* PURPOSE:  draw the clock, with tick marks and hands
*/
{  cwin << Circle(center, radius);
   int i;
   const double HOUR_TICK_LENGTH = 0.2;
   const double MINUTE_TICK_LENGTH = 0.1;
   const double HOUR_HAND_LENGTH = 0.6;
   const double MINUTE_HAND_LENGTH = 0.75;
   for (i = 0; i < 12; i++)
   {  draw_tick(i * 5, HOUR_TICK_LENGTH);
      int j;
      for (j = 1; j <= 4; j++)
         draw_tick(i * 5 + j, MINUTE_TICK_LENGTH);
   }
   draw_hand(current_time.get_minutes(), MINUTE_HAND_LENGTH);
   draw_hand((current_time.get_hours() +
      current_time.get_minutes() / 60.0) * 5, HOUR_HAND_LENGTH);
}

Player::Player()
{  level = 1;
   score = 0;
}

Player::Player(string player_name, int initial_level)
/* RECEIVES: player_name - the player name
             initial_level - the player's level (1...4)
*/            
{  name = player_name;
   level = initial_level;
   score = 0;
}

int Player::get_level() const
{  return level;
}

string Player::get_name() const
{  return name;
}

void Player::increment_score()
/* PURPOSE:  increment score
   REMARKS:  moves to next level if current level complete
*/
{  score++;
   if (score % 5 == 0 and level < 4)
      level++;
}

Game::Game()
{  rand_seed();
}

void Game::play()
/* PURPOSE:  plays the game while the player wants to continue
*/
{  read_player_information();
   string response;
   do
   {  play_round();
      response = cwin.get_string("Do you want to play again? (y/n)");
   } while (response == "y");
}

void Game::read_player_information()
/* PURPOSE:  reads player name and level
*/
{  string name = cwin.get_string("What is your name?");
   int initial_level;
   do
   {  initial_level =
         cwin.get_int("At what level do you want to start? (1-4)");
   } while (initial_level < 1 or initial_level > 4);
   player = Player(name, initial_level);
}

Time Game::random_time()
/* PURPOSE:  make a random time, depending on the level
   RETURNS:  the random time
*/
{  int level = player.get_level();
   int minutes;
   if (level == 1) minutes = 0;
   else if (level == 2) minutes = 15 * rand_int(0, 3);
   else if (level == 3) minutes = 5 * rand_int(0, 11);
   else minutes = rand_int(0, 59);
   int hours = rand_int(1, 12);
   return Time(hours, minutes, 0);
}

Time Game::get_guess()
/* PURPOSE:  get a time input from the user
   RETURNS:  the time guessed by the user
*/
{  int hours;
   do
   {  hours = cwin.get_int("Please enter hours: (1-12)");
   } while (hours < 1 or hours > 12);
   int minutes;
   do
   {  minutes = cwin.get_int("Please enter minutes: (0-59)");
   } while (minutes < 0 or minutes > 59);

   return Time(hours, minutes, 0);
}

void Game::play_round()
/* PURPOSE:  Play a round, with up to two guesses
*/
{  cwin.clear();
   Time t = random_time();
   const double CLOCK_RADIUS = 5;
   Clock clock(Point(0, 0), CLOCK_RADIUS);
   clock.set_time(t);
   clock.draw();

   Time guess = get_guess();
   if (t.seconds_from(guess) != 0)
      guess = get_guess();

   string text;
   if (t.seconds_from(guess) == 0)
   {  text = "Congratulations, " + player.get_name()
         + "! That is correct.";
      player.increment_score();
   }
   else
      text = "Sorry, " + player.get_name()
         + "! That is not correct.";
   cwin << Message(Point(-CLOCK_RADIUS, CLOCK_RADIUS + 1), text);
}

int main()
{  Game clock_game;
   clock_game.play();

   return 0;
}
